import { parseBlob } from "music-metadata"
import { FastAverageColor } from "fast-average-color"

export default class TrackManifest {
	constructor(params, ctx) {
		this.params = params
		this.ctx = ctx

		this.uid = params.uid ?? params._id
		this._id = params._id

		if (typeof params.service !== "undefined") {
			this.service = params.service
		}

		if (typeof params.overrides !== "undefined") {
			this.overrides = params.overrides
		}

		if (typeof params.cover !== "undefined") {
			this.cover = params.cover
		}

		if (typeof params.title !== "undefined") {
			this.title = params.title
		}

		if (typeof params.album !== "undefined") {
			this.album = params.album
		}

		if (typeof params.artist !== "undefined") {
			this.artist = params.artist
		}

		if (typeof params.source !== "undefined") {
			this.source = params.source
		}

		if (typeof params.metadata !== "undefined") {
			this.metadata = params.metadata
		}

		if (typeof params.liked !== "undefined") {
			this.liked = params.liked
		}

		return this
	}

	_id = null // used for api requests
	uid = null // used for internal

	title = "Untitled"
	album = "Unknown"
	artist = "Unknown"
	cover = null // set default cover url
	source = null
	metadata = {}

	// set default service to default
	service = "default"

	async initialize() {
		if (!this.params.file) {
			return this
		}

		const analyzedMetadata = await parseBlob(this.params.file, {
			skipPostHeaders: true,
		}).catch(() => ({}))

		if (analyzedMetadata.format) {
			this.metadata.format = analyzedMetadata.format.codec
		}

		if (analyzedMetadata.common) {
			this.title = analyzedMetadata.common.title ?? this.title
			this.artist = analyzedMetadata.common.artist ?? this.artist
			this.album = analyzedMetadata.common.album ?? this.album
		}

		if (analyzedMetadata.common.picture) {
			const cover = analyzedMetadata.common.picture[0]

			this._coverBlob = new Blob([cover.data], { type: cover.format })
			this.cover = URL.createObjectURL(this._coverBlob)
		}

		return this
	}

	analyzeCoverColor = async () => {
		const fac = new FastAverageColor()

		return await fac.getColorAsync(this.cover)
	}

	serviceOperations = {
		fetchLyrics: async () => {
			if (!this._id) {
				return null
			}

			const result = await this.ctx.serviceProviders.operation(
				"resolveLyrics",
				this.service,
				this,
			)

			if (this.overrides) {
				return {
					...result,
					...this.overrides,
				}
			}

			return result
		},
		fetchOverride: async () => {
			if (!this._id) {
				return null
			}

			return await this.ctx.serviceProviders.operation(
				"resolveOverride",
				this.service,
				this,
			)
		},
		toggleItemFavorite: async (to) => {
			if (!this._id) {
				return null
			}

			return await this.ctx.serviceProviders.operation(
				"toggleItemFavorite",
				this.service,
				this,
				"tracks",
				to,
			)
		},
		isItemFavorited: async () => {
			if (!this._id) {
				return null
			}

			return await this.ctx.serviceProviders.operation(
				"isItemFavorited",
				this.service,
				this,
				"tracks",
			)
		},
	}

	toSeriableObject = () => {
		return {
			_id: this._id,
			uid: this.uid,
			cover: this.cover,
			title: this.title,
			album: this.album,
			artist: this.artist,
			source: this.source,
			metadata: this.metadata,
			liked: this.liked,
		}
	}
}
